// Copyright 2022 The casbin Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package constant

const (
	ActionIndex   = "act"
	DomainIndex   = "dom"
	SubjectIndex  = "sub"
	ObjectIndex   = "obj"
	PriorityIndex = "priority"
)

const (
	AllowOverrideEffect   = "some(where (p_eft == allow))"
	DenyOverrideEffect    = "!some(where (p_eft == deny))"
	AllowAndDenyEffect    = "some(where (p_eft == allow)) && !some(where (p_eft == deny))"
	PriorityEffect        = "priority(p_eft) || deny"
	SubjectPriorityEffect = "subjectPriority(p_eft) || deny"
)
